@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2008  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -23,12 +23,16 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
 #include <stdlib.h>
 #include <string.h>
-#include <dbus/dbus.h>
 
-#include <connman/log.h>
+#include <gdbus.h>
+
+#define CONNMAN_API_SUBJECT_TO_CHANGE
+#include <connman/device.h>
 #include <connman/dbus.h>
+#include <connman/log.h>
 
 #include "inet.h"
 #include "supplicant.h"
@@ -39,21 +43,148 @@
 #define IEEE80211_CAP_IBSS      0x0002
 #define IEEE80211_CAP_PRIVACY   0x0010
 
+#define SUPPLICANT_NAME  "fi.epitest.hostap.WPASupplicant"
+#define SUPPLICANT_INTF  "fi.epitest.hostap.WPASupplicant"
+#define SUPPLICANT_PATH  "/fi/epitest/hostap/WPASupplicant"
+
+/* Taken from "WPA Supplicant - Common definitions" */
+enum supplicant_state {
+	/**
+	 * WPA_DISCONNECTED - Disconnected state
+	 *
+	 * This state indicates that client is not associated, but is likely to
+	 * start looking for an access point. This state is entered when a
+	 * connection is lost.
+	 */
+	WPA_DISCONNECTED,
+
+	/**
+	 * WPA_INACTIVE - Inactive state (wpa_supplicant disabled)
+	 *
+	 * This state is entered if there are no enabled networks in the
+	 * configuration. wpa_supplicant is not trying to associate with a new
+	 * network and external interaction (e.g., ctrl_iface call to add or
+	 * enable a network) is needed to start association.
+	 */
+	WPA_INACTIVE,
+
+	/**
+	 * WPA_SCANNING - Scanning for a network
+	 *
+	 * This state is entered when wpa_supplicant starts scanning for a
+	 * network.
+	 */
+	WPA_SCANNING,
+
+	/**
+	 * WPA_ASSOCIATING - Trying to associate with a BSS/SSID
+	 *
+	 * This state is entered when wpa_supplicant has found a suitable BSS
+	 * to associate with and the driver is configured to try to associate
+	 * with this BSS in ap_scan=1 mode. When using ap_scan=2 mode, this
+	 * state is entered when the driver is configured to try to associate
+	 * with a network using the configured SSID and security policy.
+	 */
+	WPA_ASSOCIATING,
+
+	/**
+	 * WPA_ASSOCIATED - Association completed
+	 *
+	 * This state is entered when the driver reports that association has
+	 * been successfully completed with an AP. If IEEE 802.1X is used
+	 * (with or without WPA/WPA2), wpa_supplicant remains in this state
+	 * until the IEEE 802.1X/EAPOL authentication has been completed.
+	 */
+	WPA_ASSOCIATED,
+
+	/**
+	 * WPA_4WAY_HANDSHAKE - WPA 4-Way Key Handshake in progress
+	 *
+	 * This state is entered when WPA/WPA2 4-Way Handshake is started. In
+	 * case of WPA-PSK, this happens when receiving the first EAPOL-Key
+	 * frame after association. In case of WPA-EAP, this state is entered
+	 * when the IEEE 802.1X/EAPOL authentication has been completed.
+	 */
+	WPA_4WAY_HANDSHAKE,
+
+	/**
+	 * WPA_GROUP_HANDSHAKE - WPA Group Key Handshake in progress
+	 *
+	 * This state is entered when 4-Way Key Handshake has been completed
+	 * (i.e., when the supplicant sends out message 4/4) and when Group
+	 * Key rekeying is started by the AP (i.e., when supplicant receives
+	 * message 1/2).
+	 */
+	WPA_GROUP_HANDSHAKE,
+
+	/**
+	 * WPA_COMPLETED - All authentication completed
+	 *
+	 * This state is entered when the full authentication process is
+	 * completed. In case of WPA2, this happens when the 4-Way Handshake is
+	 * successfully completed. With WPA, this state is entered after the
+	 * Group Key Handshake; with IEEE 802.1X (non-WPA) connection is
+	 * completed after dynamic keys are received (or if not used, after
+	 * the EAP authentication has been completed). With static WEP keys and
+	 * plaintext connections, this state is entered when an association
+	 * has been completed.
+	 *
+	 * This state indicates that the supplicant has completed its
+	 * processing for the association phase and that data connection is
+	 * fully configured.
+	 */
+	WPA_COMPLETED,
+
+	/**
+	 * WPA_INVALID - Invalid state (parsing error)
+	 *
+	 * This state is returned if the string input is invalid. It is not
+	 * an official wpa_supplicant state.
+	 */
+	WPA_INVALID,
+};
+
+struct supplicant_result {
+	char *identifier;
+	unsigned char *ssid;
+	unsigned int ssid_len;
+	dbus_uint16_t capabilities;
+	gboolean adhoc;
+	gboolean has_wep;
+	gboolean has_wpa;
+	gboolean has_rsn;
+	dbus_int32_t quality;
+	dbus_int32_t noise;
+	dbus_int32_t level;
+	dbus_int32_t maxrate;
+};
+
 struct supplicant_task {
 	int ifindex;
-	gchar *ifname;
-	struct connman_element *element;
-	struct supplicant_callback *callback;
-	gchar *path;
+	char *ifname;
+	struct connman_device *device;
+	struct connman_network *network;
+	char *path;
+	char *netpath;
 	gboolean created;
-	gchar *network;
 	enum supplicant_state state;
+	gboolean noscan;
+	GSList *scan_results;
 };
 
 static GSList *task_list = NULL;
 
 static DBusConnection *connection;
 
+static void free_task(struct supplicant_task *task)
+{
+	DBG("task %p", task);
+
+	g_free(task->ifname);
+	g_free(task->path);
+	g_free(task);
+}
+
 static struct supplicant_task *find_task_by_index(int index)
 {
 	GSList *list;
@@ -82,14 +213,122 @@ static struct supplicant_task *find_task_by_path(const char *path)
 	return NULL;
 }
 
-static int get_interface(struct supplicant_task *task)
+static void add_interface_reply(DBusPendingCall *call, void *user_data)
 {
-	DBusMessage *message, *reply;
+	struct supplicant_task *task = user_data;
+	DBusMessage *reply;
 	DBusError error;
 	const char *path;
 
 	DBG("task %p", task);
 
+	reply = dbus_pending_call_steal_reply(call);
+	if (reply == NULL)
+		return;
+
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
+		goto done;
+
+	dbus_error_init(&error);
+
+	if (dbus_message_get_args(reply, &error, DBUS_TYPE_OBJECT_PATH, &path,
+						DBUS_TYPE_INVALID) == FALSE) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Wrong arguments for add interface");
+		goto done;
+	}
+
+	DBG("path %s", path);
+
+	task->path = g_strdup(path);
+	task->created = TRUE;
+
+	connman_device_set_powered(task->device, TRUE);
+
+done:
+	dbus_message_unref(reply);
+}
+
+static int add_interface(struct supplicant_task *task)
+{
+	DBusMessage *message;
+	DBusPendingCall *call;
+
+	DBG("task %p", task);
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, SUPPLICANT_PATH,
+					SUPPLICANT_INTF, "addInterface");
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_append_args(message, DBUS_TYPE_STRING, &task->ifname,
+							DBUS_TYPE_INVALID);
+
+	if (dbus_connection_send_with_reply(connection, message,
+						&call, TIMEOUT) == FALSE) {
+		connman_error("Failed to add interface");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_pending_call_set_notify(call, add_interface_reply, task, NULL);
+
+	dbus_message_unref(message);
+
+	return -EINPROGRESS;
+}
+
+static void get_interface_reply(DBusPendingCall *call, void *user_data)
+{
+	struct supplicant_task *task = user_data;
+	DBusMessage *reply;
+	DBusError error;
+	const char *path;
+
+	DBG("task %p", task);
+
+	reply = dbus_pending_call_steal_reply(call);
+	if (reply == NULL)
+		return;
+
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
+		add_interface(task);
+		goto done;
+	}
+
+	dbus_error_init(&error);
+
+	if (dbus_message_get_args(reply, &error, DBUS_TYPE_OBJECT_PATH, &path,
+						DBUS_TYPE_INVALID) == FALSE) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Wrong arguments for get interface");
+		goto done;
+	}
+
+	DBG("path %s", path);
+
+	task->path = g_strdup(path);
+	task->created = FALSE;
+
+	connman_device_set_powered(task->device, TRUE);
+
+done:
+	dbus_message_unref(reply);
+}
+
+static int create_interface(struct supplicant_task *task)
+{
+	DBusMessage *message;
+	DBusPendingCall *call;
+
+	DBG("task %p", task);
+
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, SUPPLICANT_PATH,
 					SUPPLICANT_INTF, "getInterface");
 	if (message == NULL)
@@ -98,109 +337,47 @@ static int get_interface(struct supplicant_task *task)
 	dbus_message_append_args(message, DBUS_TYPE_STRING, &task->ifname,
 							DBUS_TYPE_INVALID);
 
-	dbus_error_init(&error);
-
-	reply = dbus_connection_send_with_reply_and_block(connection,
-							message, -1, &error);
-	if (reply == NULL) {
-		if (dbus_error_is_set(&error) == TRUE) {
-			connman_error("%s", error.message);
-			dbus_error_free(&error);
-		} else
-			connman_error("Failed to get interface");
+	if (dbus_connection_send_with_reply(connection, message,
+						&call, TIMEOUT) == FALSE) {
+		connman_error("Failed to get interface");
 		dbus_message_unref(message);
 		return -EIO;
 	}
 
+	dbus_pending_call_set_notify(call, get_interface_reply, task, NULL);
+
 	dbus_message_unref(message);
 
-	dbus_error_init(&error);
-
-	if (dbus_message_get_args(reply, &error, DBUS_TYPE_OBJECT_PATH, &path,
-						DBUS_TYPE_INVALID) == FALSE) {
-		if (dbus_error_is_set(&error) == TRUE) {
-			connman_error("%s", error.message);
-			dbus_error_free(&error);
-		} else
-			connman_error("Wrong arguments for interface");
-		dbus_message_unref(reply);
-		return -EIO;
-	}
-
-	DBG("path %s", path);
-
-	task->path = g_strdup(path);
-	task->created = FALSE;
-
-	dbus_message_unref(reply);
-
-	return 0;
+	return -EINPROGRESS;
 }
 
-static int add_interface(struct supplicant_task *task)
+static void remove_interface_reply(DBusPendingCall *call, void *user_data)
 {
-	DBusMessage *message, *reply;
-	DBusError error;
-	const char *path;
+	struct supplicant_task *task = user_data;
+	DBusMessage *reply;
 
 	DBG("task %p", task);
 
-	message = dbus_message_new_method_call(SUPPLICANT_NAME, SUPPLICANT_PATH,
-					SUPPLICANT_INTF, "addInterface");
-	if (message == NULL)
-		return -ENOMEM;
+	reply = dbus_pending_call_steal_reply(call);
 
-	dbus_error_init(&error);
+	connman_device_set_powered(task->device, FALSE);
 
-	dbus_message_append_args(message, DBUS_TYPE_STRING, &task->ifname,
-							DBUS_TYPE_INVALID);
-
-	reply = dbus_connection_send_with_reply_and_block(connection,
-							message, -1, &error);
-	if (reply == NULL) {
-		if (dbus_error_is_set(&error) == TRUE) {
-			connman_error("%s", error.message);
-			dbus_error_free(&error);
-		} else
-			connman_error("Failed to add interface");
-		dbus_message_unref(message);
-		return -EIO;
-	}
-
-	dbus_message_unref(message);
-
-	dbus_error_init(&error);
-
-	if (dbus_message_get_args(reply, &error, DBUS_TYPE_OBJECT_PATH, &path,
-						DBUS_TYPE_INVALID) == FALSE) {
-		if (dbus_error_is_set(&error) == TRUE) {
-			connman_error("%s", error.message);
-			dbus_error_free(&error);
-		} else
-			connman_error("Wrong arguments for interface");
-		dbus_message_unref(reply);
-		return -EIO;
-	}
-
-	DBG("path %s", path);
-
-	task->path = g_strdup(path);
-	task->created = TRUE;
+	free_task(task);
 
 	dbus_message_unref(reply);
-
-	return 0;
 }
 
 static int remove_interface(struct supplicant_task *task)
 {
-	DBusMessage *message, *reply;
-	DBusError error;
+	DBusMessage *message;
+	DBusPendingCall *call;
 
 	DBG("task %p", task);
 
-	if (task->created == FALSE)
-		return -EINVAL;
+	if (task->created == FALSE) {
+		connman_device_set_powered(task->device, FALSE);
+		return 0;
+	}
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, SUPPLICANT_PATH,
 					SUPPLICANT_INTF, "removeInterface");
@@ -210,27 +387,21 @@ static int remove_interface(struct supplicant_task *task)
 	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->path,
 							DBUS_TYPE_INVALID);
 
-	dbus_error_init(&error);
-
-	reply = dbus_connection_send_with_reply_and_block(connection,
-							message, -1, &error);
-	if (reply == NULL) {
-		if (dbus_error_is_set(&error) == TRUE) {
-			connman_error("%s", error.message);
-			dbus_error_free(&error);
-		} else
-			connman_error("Failed to remove interface");
+	if (dbus_connection_send_with_reply(connection, message,
+						&call, TIMEOUT) == FALSE) {
+		connman_error("Failed to remove interface");
 		dbus_message_unref(message);
 		return -EIO;
 	}
 
+	dbus_pending_call_set_notify(call, remove_interface_reply, task, NULL);
+
 	dbus_message_unref(message);
 
-	dbus_message_unref(reply);
-
-	return 0;
+	return -EINPROGRESS;
 }
 
+#if 0
 static int set_ap_scan(struct supplicant_task *task)
 {
 	DBusMessage *message, *reply;
@@ -267,6 +438,7 @@ static int set_ap_scan(struct supplicant_task *task)
 
 	return 0;
 }
+#endif
 
 static int add_network(struct supplicant_task *task)
 {
@@ -276,7 +448,7 @@ static int add_network(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
-	if (task->network != NULL)
+	if (task->netpath != NULL)
 		return -EALREADY;
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
@@ -315,7 +487,7 @@ static int add_network(struct supplicant_task *task)
 
 	DBG("path %s", path);
 
-	task->network = g_strdup(path);
+	task->netpath = g_strdup(path);
 
 	dbus_message_unref(reply);
 
@@ -329,7 +501,7 @@ static int remove_network(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
-	if (task->network == NULL)
+	if (task->netpath == NULL)
 		return -EINVAL;
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
@@ -337,7 +509,7 @@ static int remove_network(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
-	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->network,
+	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->netpath,
 							DBUS_TYPE_INVALID);
 
 	dbus_error_init(&error);
@@ -358,8 +530,8 @@ static int remove_network(struct supplicant_task *task)
 
 	dbus_message_unref(reply);
 
-	g_free(task->network);
-	task->network = NULL;
+	g_free(task->netpath);
+	task->netpath = NULL;
 
 	return 0;
 }
@@ -371,7 +543,7 @@ static int select_network(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
-	if (task->network == NULL)
+	if (task->netpath == NULL)
 		return -EINVAL;
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
@@ -379,7 +551,7 @@ static int select_network(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
-	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->network,
+	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->netpath,
 							DBUS_TYPE_INVALID);
 
 	dbus_error_init(&error);
@@ -410,10 +582,10 @@ static int enable_network(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
-	if (task->network == NULL)
+	if (task->netpath == NULL)
 		return -EINVAL;
 
-	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->network,
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->netpath,
 					SUPPLICANT_INTF ".Network", "enable");
 	if (message == NULL)
 		return -ENOMEM;
@@ -446,10 +618,10 @@ static int disable_network(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
-	if (task->network == NULL)
+	if (task->netpath == NULL)
 		return -EINVAL;
 
-	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->network,
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->netpath,
 					SUPPLICANT_INTF ".Network", "disable");
 	if (message == NULL)
 		return -ENOMEM;
@@ -485,10 +657,10 @@ static int set_network(struct supplicant_task *task,
 
 	DBG("task %p", task);
 
-	if (task->network == NULL)
+	if (task->netpath == NULL)
 		return -EINVAL;
 
-	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->network,
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->netpath,
 					SUPPLICANT_INTF ".Network", "set");
 	if (message == NULL)
 		return -ENOMEM;
@@ -531,7 +703,7 @@ static int set_network(struct supplicant_task *task,
 					key[i] = (unsigned char) strtol(tmp,
 								NULL, 16);
 				}
-			        connman_dbus_dict_append_array(&dict,
+				connman_dbus_dict_append_array(&dict,
 						"wep_key0", DBUS_TYPE_BYTE,
 							&key, size / 2);
 				free(key);
@@ -595,8 +767,8 @@ static int initiate_scan(struct supplicant_task *task)
 	return 0;
 }
 
-static void extract_ssid(struct supplicant_network *network,
-						DBusMessageIter *value)
+static void extract_ssid(DBusMessageIter *value,
+					struct supplicant_result *result)
 {
 	DBusMessageIter array;
 	unsigned char *ssid;
@@ -608,22 +780,22 @@ static void extract_ssid(struct supplicant_network *network,
 	if (ssid_len < 1)
 		return;
 
-	network->ssid = g_try_malloc(ssid_len);
-	if (network->ssid == NULL)
+	result->ssid = g_try_malloc(ssid_len);
+	if (result->ssid == NULL)
 		return;
 
-	memcpy(network->ssid, ssid, ssid_len);
-	network->ssid_len = ssid_len;
+	memcpy(result->ssid, ssid, ssid_len);
+	result->ssid_len = ssid_len;
 
-	network->identifier = g_try_malloc0(ssid_len + 1);
-	if (network->identifier == NULL)
+	result->identifier = g_try_malloc0(ssid_len + 1);
+	if (result->identifier == NULL)
 		return;
 
-	memcpy(network->identifier, ssid, ssid_len);
+	memcpy(result->identifier, ssid, ssid_len);
 }
 
-static void extract_wpaie(struct supplicant_network *network,
-						DBusMessageIter *value)
+static void extract_wpaie(DBusMessageIter *value,
+					struct supplicant_result *result)
 {
 	DBusMessageIter array;
 	unsigned char *ie;
@@ -633,11 +805,11 @@ static void extract_wpaie(struct supplicant_network *network,
 	dbus_message_iter_get_fixed_array(&array, &ie, &ie_len);
 
 	if (ie_len > 0)
-		network->has_wpa = TRUE;
+		result->has_wpa = TRUE;
 }
 
-static void extract_rsnie(struct supplicant_network *network,
-						DBusMessageIter *value)
+static void extract_rsnie(DBusMessageIter *value,
+					struct supplicant_result *result)
 {
 	DBusMessageIter array;
 	unsigned char *ie;
@@ -647,32 +819,51 @@ static void extract_rsnie(struct supplicant_network *network,
 	dbus_message_iter_get_fixed_array(&array, &ie, &ie_len);
 
 	if (ie_len > 0)
-		network->has_rsn = TRUE;
+		result->has_rsn = TRUE;
 }
 
-static void extract_capabilites(struct supplicant_network *network,
-						DBusMessageIter *value)
+static void extract_capabilites(DBusMessageIter *value,
+					struct supplicant_result *result)
 {
-	dbus_message_iter_get_basic(value, &network->capabilities);
+	dbus_message_iter_get_basic(value, &result->capabilities);
 
-	if (network->capabilities & IEEE80211_CAP_PRIVACY)
-		network->has_wep = TRUE;
+	if (result->capabilities & IEEE80211_CAP_ESS)
+		result->adhoc = FALSE;
+	else if (result->capabilities & IEEE80211_CAP_IBSS)
+		result->adhoc = TRUE;
+
+	if (result->capabilities & IEEE80211_CAP_PRIVACY)
+		result->has_wep = TRUE;
 }
 
+static void get_properties(struct supplicant_task *task);
+
 static void properties_reply(DBusPendingCall *call, void *user_data)
 {
 	struct supplicant_task *task = user_data;
-	struct supplicant_network *network;
+	struct supplicant_result result;
+	struct connman_network *network;
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
+	char *security, *temp = NULL;
+	unsigned char strength;
+	unsigned int i;
 
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
+	if (reply == NULL) {
+		get_properties(task);
+		return;
+	}
 
-	network = g_try_new0(struct supplicant_network, 1);
-	if (network == NULL)
-		goto done;
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
+		dbus_message_unref(reply);
+		get_properties(task);
+		return;
+	}
+
+	memset(&result, 0, sizeof(result));
 
 	dbus_message_iter_init(reply, &array);
 
@@ -706,61 +897,137 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 		 */
 
 		if (g_str_equal(key, "ssid") == TRUE)
-			extract_ssid(network, &value);
+			extract_ssid(&value, &result);
 		else if (g_str_equal(key, "wpaie") == TRUE)
-			extract_wpaie(network, &value);
+			extract_wpaie(&value, &result);
 		else if (g_str_equal(key, "rsnie") == TRUE)
-			extract_rsnie(network, &value);
+			extract_rsnie(&value, &result);
 		else if (g_str_equal(key, "capabilities") == TRUE)
-			extract_capabilites(network, &value);
+			extract_capabilites(&value, &result);
 		else if (g_str_equal(key, "quality") == TRUE)
-			dbus_message_iter_get_basic(&value, &network->quality);
+			dbus_message_iter_get_basic(&value, &result.quality);
 		else if (g_str_equal(key, "noise") == TRUE)
-			dbus_message_iter_get_basic(&value, &network->noise);
+			dbus_message_iter_get_basic(&value, &result.noise);
 		else if (g_str_equal(key, "level") == TRUE)
-			dbus_message_iter_get_basic(&value, &network->level);
+			dbus_message_iter_get_basic(&value, &result.level);
 		else if (g_str_equal(key, "maxrate") == TRUE)
-			dbus_message_iter_get_basic(&value, &network->maxrate);
-
+			dbus_message_iter_get_basic(&value, &result.maxrate);
 
 		dbus_message_iter_next(&dict);
 	}
 
-	if (task->callback && task->callback->scan_result)
-		task->callback->scan_result(task->element, network);
+	if (result.identifier == NULL)
+		goto done;
 
-	g_free(network->identifier);
-	g_free(network->ssid);
-	g_free(network);
+	if (result.identifier[0] == '\0')
+		goto done;
+
+	temp = g_strdup(result.identifier);
+	if (temp == NULL)
+		goto done;
+
+	for (i = 0; i < strlen(temp); i++) {
+		char tmp = temp[i];
+		if ((tmp < '0' || tmp > '9') && (tmp < 'A' || tmp > 'Z') &&
+						(tmp < 'a' || tmp > 'z'))
+			temp[i] = '_';
+	}
+
+	strength = result.quality;
+
+	if (result.has_rsn == TRUE)
+		security = "wpa2";
+	else if (result.has_wpa == TRUE)
+		security = "wpa";
+	else if (result.has_wep == TRUE)
+		security = "wep";
+	else
+		security = "none";
+
+	network = connman_device_get_network(task->device, temp);
+	if (network == NULL) {
+		const char *mode;
+		int index;
+
+		network = connman_network_create(temp,
+						CONNMAN_NETWORK_TYPE_WIFI);
+		if (network == NULL)
+			goto done;
+
+		index = connman_device_get_index(task->device);
+		connman_network_set_index(network, index);
+
+		connman_network_set_protocol(network,
+						CONNMAN_NETWORK_PROTOCOL_IP);
+
+		connman_network_set_string(network, "Name", result.identifier);
+
+		connman_network_set_blob(network, "WiFi.SSID",
+						result.ssid, result.ssid_len);
+
+		mode = (result.adhoc == TRUE) ? "adhoc" : "managed";
+		connman_network_set_string(network, "WiFi.Mode", mode);
+
+		DBG("%s (%s %s) strength %d", result.identifier, mode,
+							security, strength);
+
+		if (connman_device_add_network(task->device, network) < 0) {
+			connman_network_unref(network);
+			goto done;
+		}
+	}
+
+	connman_network_set_available(network, TRUE);
+	connman_network_set_uint8(network, "Strength", strength);
+
+	connman_network_set_string(network, "WiFi.Security", security);
 
 done:
+	g_free(result.identifier);
+	g_free(result.ssid);
+	g_free(temp);
+
 	dbus_message_unref(reply);
+
+	get_properties(task);
 }
 
-static int get_network_properties(struct supplicant_task *task,
-							const char *path)
+static void get_properties(struct supplicant_task *task)
 {
 	DBusMessage *message;
 	DBusPendingCall *call;
+	char *path;
+
+	path = g_slist_nth_data(task->scan_results, 0);
+	if (path == NULL)
+		goto noscan;
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, path,
 						SUPPLICANT_INTF ".BSSID",
 								"properties");
+
+	task->scan_results = g_slist_remove(task->scan_results, path);
+	g_free(path);
+
 	if (message == NULL)
-		return -ENOMEM;
+		goto noscan;
 
 	if (dbus_connection_send_with_reply(connection, message,
 						&call, TIMEOUT) == FALSE) {
 		connman_error("Failed to get network properties");
 		dbus_message_unref(message);
-		return -EIO;
+		goto noscan;
 	}
 
 	dbus_pending_call_set_notify(call, properties_reply, task, NULL);
 
 	dbus_message_unref(message);
 
-	return 0;
+	return;
+
+noscan:
+	if (task->noscan == FALSE)
+		connman_device_set_scanning(task->device, FALSE);
 }
 
 static void scan_results_reply(DBusPendingCall *call, void *user_data)
@@ -774,6 +1041,11 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
+	if (reply == NULL)
+		goto noscan;
+
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
+		goto done;
 
 	dbus_error_init(&error);
 
@@ -789,19 +1061,34 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 		goto done;
 	}
 
-	if (task->callback && task->callback->clear_results)
-			task->callback->clear_results(task->element);
+	if (num_results == 0)
+		goto done;
 
-	for (i = 0; i < num_results; i++)
-		get_network_properties(task, results[i]);
+	for (i = 0; i < num_results; i++) {
+		char *path = g_strdup(results[i]);
+		if (path == NULL)
+			continue;
+
+		task->scan_results = g_slist_append(task->scan_results, path);
+	}
 
 	g_strfreev(results);
 
+	dbus_message_unref(reply);
+
+	get_properties(task);
+
+	return;
+
 done:
 	dbus_message_unref(reply);
+
+noscan:
+	if (task->noscan == FALSE)
+		connman_device_set_scanning(task->device, FALSE);
 }
 
-static int scan_results_available(struct supplicant_task *task)
+static void scan_results_available(struct supplicant_task *task)
 {
 	DBusMessage *message;
 	DBusPendingCall *call;
@@ -812,31 +1099,55 @@ static int scan_results_available(struct supplicant_task *task)
 						SUPPLICANT_INTF ".Interface",
 							"scanResults");
 	if (message == NULL)
-		return -ENOMEM;
+		return;
 
 	if (dbus_connection_send_with_reply(connection, message,
 						&call, TIMEOUT) == FALSE) {
 		connman_error("Failed to request scan result");
-		dbus_message_unref(message);
-		return -EIO;
+		goto done;
 	}
 
+	if (task->noscan == FALSE)
+		connman_device_set_scanning(task->device, TRUE);
+
 	dbus_pending_call_set_notify(call, scan_results_reply, task, NULL);
 
+done:
 	dbus_message_unref(message);
+}
 
-	return 0;
+static enum supplicant_state string2state(const char *state)
+{
+	if (g_str_equal(state, "INACTIVE") == TRUE)
+		return WPA_INACTIVE;
+	else if (g_str_equal(state, "SCANNING") == TRUE)
+		return WPA_SCANNING;
+	else if (g_str_equal(state, "ASSOCIATING") == TRUE)
+		return WPA_ASSOCIATING;
+	else if (g_str_equal(state, "ASSOCIATED") == TRUE)
+		return WPA_ASSOCIATED;
+	else if (g_str_equal(state, "GROUP_HANDSHAKE") == TRUE)
+		return WPA_GROUP_HANDSHAKE;
+	else if (g_str_equal(state, "4WAY_HANDSHAKE") == TRUE)
+		return WPA_4WAY_HANDSHAKE;
+	else if (g_str_equal(state, "COMPLETED") == TRUE)
+		return WPA_COMPLETED;
+	else if (g_str_equal(state, "DISCONNECTED") == TRUE)
+		return WPA_DISCONNECTED;
+	else
+		return WPA_INVALID;
 }
 
 static void state_change(struct supplicant_task *task, DBusMessage *msg)
 {
 	DBusError error;
-	const char *state, *previous;
+	const char *newstate, *oldstate;
+	enum supplicant_state state;
 
 	dbus_error_init(&error);
 
-	if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &state,
-						DBUS_TYPE_STRING, &previous,
+	if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &newstate,
+						DBUS_TYPE_STRING, &oldstate,
 						DBUS_TYPE_INVALID) == FALSE) {
 		if (dbus_error_is_set(&error) == TRUE) {
 			connman_error("%s", error.message);
@@ -846,34 +1157,50 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		return;
 	}
 
-	DBG("state %s ==> %s", previous, state);
+	DBG("state %s ==> %s", oldstate, newstate);
 
-	if (g_str_equal(state, "INACTIVE") == TRUE)
-		task->state = STATE_INACTIVE;
-	else if (g_str_equal(state, "SCANNING") == TRUE)
-		task->state = STATE_SCANNING;
-	else if (g_str_equal(state, "ASSOCIATING") == TRUE)
-		task->state = STATE_ASSOCIATING;
-	else if (g_str_equal(state, "ASSOCIATED") == TRUE)
-		task->state = STATE_ASSOCIATED;
-	else if (g_str_equal(state, "GROUP_HANDSHAKE") == TRUE)
-		task->state = STATE_4WAY_HANDSHAKE;
-	else if (g_str_equal(state, "4WAY_HANDSHAKE") == TRUE)
-		task->state = STATE_4WAY_HANDSHAKE;
-	else if (g_str_equal(state, "COMPLETED") == TRUE)
-		task->state = STATE_COMPLETED;
-	else if (g_str_equal(state, "DISCONNECTED") == TRUE)
-		task->state = STATE_DISCONNECTED;
+	state = string2state(newstate);
+	if (state == WPA_INVALID)
+		return;
 
-	if (task->callback && task->callback->state_change)
-		task->callback->state_change(task->element, task->state);
+	task->state = state;
 
 	switch (task->state) {
-	case STATE_COMPLETED:
-		/* carrier on */
+	case WPA_SCANNING:
+		task->noscan = TRUE;
+		connman_device_set_scanning(task->device, TRUE);
 		break;
-	case STATE_DISCONNECTED:
+	case WPA_ASSOCIATING:
+	case WPA_ASSOCIATED:
+	case WPA_4WAY_HANDSHAKE:
+	case WPA_GROUP_HANDSHAKE:
+		task->noscan = TRUE;
+		break;
+	case WPA_COMPLETED:
+	case WPA_DISCONNECTED:
+		task->noscan = FALSE;
+		break;
+	case WPA_INACTIVE:
+		task->noscan = FALSE;
+		connman_device_set_scanning(task->device, FALSE);
+		break;
+	case WPA_INVALID:
+		break;
+	}
+
+	if (task->network == NULL)
+		return;
+
+	switch (task->state) {
+	case WPA_COMPLETED:
+		/* carrier on */
+		connman_network_set_connected(task->network, TRUE);
+		connman_device_set_scanning(task->device, FALSE);
+		break;
+	case WPA_DISCONNECTED:
 		/* carrier off */
+		connman_network_set_connected(task->network, FALSE);
+		connman_device_set_scanning(task->device, FALSE);
 		break;
 	default:
 		break;
@@ -912,70 +1239,19 @@ static DBusHandlerResult supplicant_filter(DBusConnection *conn,
 	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
-static int add_filter(struct supplicant_task *task)
-{
-	DBusError error;
-	gchar *filter;
-
-	filter = g_strdup_printf("type=signal,interface=%s.Interface,path=%s",
-						SUPPLICANT_INTF, task->path);
-
-	DBG("filter %s", filter);
-
-	dbus_error_init(&error);
-
-	dbus_bus_add_match(connection, filter, &error);
-
-	g_free(filter);
-
-	if (dbus_error_is_set(&error) == TRUE) {
-		connman_error("Can't add match: %s", error.message);
-		dbus_error_free(&error);
-	}
-
-	return 0;
-}
-
-static int remove_filter(struct supplicant_task *task)
-{
-	DBusError error;
-	gchar *filter;
-
-	filter = g_strdup_printf("type=signal,interface=%s.Interface,path=%s",
-						SUPPLICANT_INTF, task->path);
-
-	DBG("filter %s", filter);
-
-	dbus_error_init(&error);
-
-	dbus_bus_remove_match(connection, filter, &error);
-
-	g_free(filter);
-
-	if (dbus_error_is_set(&error) == TRUE) {
-		connman_error("Can't add match: %s", error.message);
-		dbus_error_free(&error);
-	}
-
-	return 0;
-}
-
-int __supplicant_start(struct connman_element *element,
-					struct supplicant_callback *callback)
+int supplicant_start(struct connman_device *device)
 {
 	struct supplicant_task *task;
-	int err;
 
-	DBG("element %p name %s", element, element->name);
+	DBG("device %p", device);
 
 	task = g_try_new0(struct supplicant_task, 1);
 	if (task == NULL)
 		return -ENOMEM;
 
-	task->ifindex = element->index;
-	task->ifname = inet_index2name(element->index);
-	task->element = element;
-	task->callback = callback;
+	task->ifindex = connman_device_get_index(device);
+	task->ifname = inet_index2name(task->ifindex);
+	task->device = device;
 
 	if (task->ifname == NULL) {
 		g_free(task);
@@ -983,33 +1259,22 @@ int __supplicant_start(struct connman_element *element,
 	}
 
 	task->created = FALSE;
-	task->state = STATE_INACTIVE;
+	task->noscan = FALSE;
+	task->state = WPA_INVALID;
 
 	task_list = g_slist_append(task_list, task);
 
-	err = get_interface(task);
-	if (err < 0) {
-		err = add_interface(task);
-		if (err < 0) {
-			g_free(task);
-			return err;
-		}
-	}
-
-	add_filter(task);
-
-	set_ap_scan(task);
-
-	return 0;
+	return create_interface(task);
 }
 
-int __supplicant_stop(struct connman_element *element)
+int supplicant_stop(struct connman_device *device)
 {
+	int index = connman_device_get_index(device);
 	struct supplicant_task *task;
 
-	DBG("element %p name %s", element, element->name);
+	DBG("device %p", device);
 
-	task = find_task_by_index(element->index);
+	task = find_task_by_index(index);
 	if (task == NULL)
 		return -ENODEV;
 
@@ -1019,35 +1284,28 @@ int __supplicant_stop(struct connman_element *element)
 
 	remove_network(task);
 
-	remove_filter(task);
-
-	remove_interface(task);
-
-	g_free(task->ifname);
-	g_free(task->path);
-	g_free(task);
-
-	return 0;
+	return remove_interface(task);
 }
 
-int __supplicant_scan(struct connman_element *element)
+int supplicant_scan(struct connman_device *device)
 {
+	int index = connman_device_get_index(device);
 	struct supplicant_task *task;
 	int err;
 
-	DBG("element %p name %s", element, element->name);
+	DBG("device %p", device);
 
-	task = find_task_by_index(element->index);
+	task = find_task_by_index(index);
 	if (task == NULL)
 		return -ENODEV;
 
 	switch (task->state) {
-	case STATE_SCANNING:
+	case WPA_SCANNING:
 		return -EALREADY;
-	case STATE_ASSOCIATING:
-	case STATE_ASSOCIATED:
-	case STATE_4WAY_HANDSHAKE:
-	case STATE_GROUP_HANDSHAKE:
+	case WPA_ASSOCIATING:
+	case WPA_ASSOCIATED:
+	case WPA_4WAY_HANDSHAKE:
+	case WPA_GROUP_HANDSHAKE:
 		return -EBUSY;
 	default:
 		break;
@@ -1058,18 +1316,37 @@ int __supplicant_scan(struct connman_element *element)
 	return 0;
 }
 
-int __supplicant_connect(struct connman_element *element,
-				const unsigned char *ssid, int ssid_len,
-				const char *security, const char *passphrase)
+int supplicant_connect(struct connman_network *network)
 {
 	struct supplicant_task *task;
+	const char *security, *passphrase;
+	const void *ssid;
+	unsigned int ssid_len;
+	int index;
 
-	DBG("element %p name %s", element, element->name);
+	DBG("network %p", network);
 
-	task = find_task_by_index(element->index);
+	security = connman_network_get_string(network, "WiFi.Security");
+	passphrase = connman_network_get_string(network, "WiFi.Passphrase");
+
+	ssid = connman_network_get_blob(network, "WiFi.SSID", &ssid_len);
+
+	DBG("security %s passphrase %s", security, passphrase);
+
+	if (security == NULL && passphrase == NULL)
+		return -EINVAL;
+
+	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
+		return -EINVAL;
+
+	index = connman_network_get_index(network);
+
+	task = find_task_by_index(index);
 	if (task == NULL)
 		return -ENODEV;
 
+	task->network = connman_network_ref(network);
+
 	add_network(task);
 
 	select_network(task);
@@ -1082,13 +1359,16 @@ int __supplicant_connect(struct connman_element *element,
 	return 0;
 }
 
-int __supplicant_disconnect(struct connman_element *element)
+int supplicant_disconnect(struct connman_network *network)
 {
 	struct supplicant_task *task;
+	int index;
 
-	DBG("element %p name %s", element, element->name);
+	DBG("network %p", network);
 
-	task = find_task_by_index(element->index);
+	index = connman_network_get_index(network);
+
+	task = find_task_by_index(index);
 	if (task == NULL)
 		return -ENODEV;
 
@@ -1096,23 +1376,141 @@ int __supplicant_disconnect(struct connman_element *element)
 
 	remove_network(task);
 
+	connman_network_set_connected(task->network, FALSE);
+
+	connman_network_unref(task->network);
+
 	return 0;
 }
 
-int __supplicant_init(DBusConnection *conn)
+static void supplicant_activate(DBusConnection *conn)
 {
-	connection = conn;
+	DBusMessage *message;
+
+	DBG("conn %p", conn);
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, "/",
+				DBUS_INTERFACE_INTROSPECTABLE, "Introspect");
+	if (message == NULL)
+		return;
+
+	dbus_message_set_no_reply(message, TRUE);
+
+	dbus_connection_send(conn, message, NULL);
+
+	dbus_message_unref(message);
+}
+
+static GSList *driver_list = NULL;
+
+static void supplicant_probe(DBusConnection *conn, void *user_data)
+{
+	GSList *list;
+
+	DBG("conn %p", conn);
+
+	for (list = driver_list; list; list = list->next) {
+		struct supplicant_driver *driver = list->data;
+
+		DBG("driver %p name %s", driver, driver->name);
+
+		if (driver->probe)
+			driver->probe();
+	}
+}
+
+static void supplicant_remove(DBusConnection *conn, void *user_data)
+{
+	GSList *list;
+
+	DBG("conn %p", conn);
+
+	for (list = driver_list; list; list = list->next) {
+		struct supplicant_driver *driver = list->data;
+
+		DBG("driver %p name %s", driver, driver->name);
+
+		if (driver->remove)
+			driver->remove();
+	}
+}
+
+static const char *supplicant_rule = "type=signal,"
+				"interface=" SUPPLICANT_INTF ".Interface";
+static guint watch;
+
+static int supplicant_create(void)
+{
+	if (g_slist_length(driver_list) > 0)
+		return 0;
+
+	connection = connman_dbus_get_connection();
+	if (connection == NULL)
+		return -EIO;
+
+	DBG("connection %p", connection);
 
 	if (dbus_connection_add_filter(connection,
 				supplicant_filter, NULL, NULL) == FALSE) {
-		dbus_connection_unref(connection);
+		connection = connman_dbus_get_connection();
 		return -EIO;
 	}
 
+	dbus_bus_add_match(connection, supplicant_rule, NULL);
+	dbus_connection_flush(connection);
+
+	watch = g_dbus_add_service_watch(connection, SUPPLICANT_NAME,
+			supplicant_probe, supplicant_remove, NULL, NULL);
+
 	return 0;
 }
 
-void __supplicant_exit(void)
+static void supplicant_destroy(void)
 {
+	if (g_slist_length(driver_list) > 0)
+		return;
+
+	DBG("connection %p", connection);
+
+	if (watch > 0)
+		g_dbus_remove_watch(connection, watch);
+
+	dbus_bus_remove_match(connection, supplicant_rule, NULL);
+	dbus_connection_flush(connection);
+
 	dbus_connection_remove_filter(connection, supplicant_filter, NULL);
+
+	dbus_connection_unref(connection);
+	connection = NULL;
+}
+
+int supplicant_register(struct supplicant_driver *driver)
+{
+	int err;
+
+	DBG("driver %p name %s", driver, driver->name);
+
+	err = supplicant_create();
+	if (err < 0)
+		return err;
+
+	driver_list = g_slist_append(driver_list, driver);
+
+	if (g_dbus_check_service(connection, SUPPLICANT_NAME) == TRUE)
+		supplicant_probe(connection, NULL);
+	else
+		supplicant_activate(connection);
+
+	return 0;
+}
+
+void supplicant_unregister(struct supplicant_driver *driver)
+{
+	DBG("driver %p name %s", driver, driver->name);
+
+	supplicant_remove(connection, NULL);
+
+	driver_list = g_slist_remove(driver_list, driver);
+
+	supplicant_destroy();
 }
